# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from itertools import chain
from json import dumps, loads
from urllib.parse import parse_qs, unquote
from xml.sax.saxutils import escape

from flask import Flask, Response, request

from rucio.api.quarantined_replica import quarantine_file_replicas
from rucio.api.replica import add_replicas, list_replicas, list_dataset_replicas, list_dataset_replicas_bulk, \
    delete_replicas, get_did_from_pfns, update_replicas_states, declare_bad_file_replicas, add_bad_dids, add_bad_pfns, \
    get_suspicious_files, declare_suspicious_file_replicas, list_bad_replicas_status, get_bad_replicas_summary, \
    list_datasets_per_rse, set_tombstone, list_dataset_replicas_vp
from rucio.common.config import config_get, config_get_int
from rucio.common.constants import SUPPORTED_PROTOCOLS
from rucio.common.exception import AccessDenied, DataIdentifierAlreadyExists, InvalidType, DataIdentifierNotFound, \
    Duplicate, InvalidPath, ResourceTemporaryUnavailable, RSENotFound, ReplicaNotFound, InvalidObject, ScopeNotFound, ReplicaIsLocked
from rucio.common.utils import parse_response, APIEncoder, render_json_list
from rucio.core.replica_sorter import sort_replicas
from rucio.db.sqla.constants import BadFilesStatus
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import check_accept_header_wrapper_flask, try_stream, parse_scope_name, \
    response_headers, generate_http_error_flask, ErrorHandlingMethodView, json_parameters, param_get


def _sorted_with_priorities(replicas, sorted_pfns, limit=None):
    """
    Pick up to "limit" replicas from "replicas" in the order given by sorted_pfns.
    Sets the corresponding priority in returned replicas.

    :param replicas: Dictionary {pfn: replica_definition}.
    :param sorted_pfns: Sorted list of pfns.
    :param limit: only return this many replicas
    :yields: index and corresponding pfn
    """
    for idx, pfn in enumerate(sorted_pfns, start=1):
        if limit is None or idx <= limit:
            replica = replicas[pfn]
            replica['priority'] = idx
            yield pfn, replica


def _generate_one_metalink_file(rfile, policy_schema, detailed_url=True):
    yield ' <file name="' + rfile['name'] + '">\n'

    if 'parents' in rfile and rfile['parents']:
        yield '  <parents>\n'
        for parent in rfile['parents']:
            yield '   <did>' + parent + '</did>\n'
        yield '  </parents>\n'

    yield '  <identity>' + rfile['scope'] + ':' + rfile['name'] + '</identity>\n'
    if rfile['adler32'] is not None:
        yield '  <hash type="adler32">' + rfile['adler32'] + '</hash>\n'
    if rfile['md5'] is not None:
        yield '  <hash type="md5">' + rfile['md5'] + '</hash>\n'

    yield '  <size>' + str(rfile['bytes']) + '</size>\n'

    yield f'  <glfn name="/{policy_schema}/rucio/{rfile["scope"]}:{rfile["name"]}"></glfn>\n'

    for pfn, replica in rfile['pfns'].items():
        if detailed_url:
            yield (
                '  '
                f'<url location="{replica["rse"]}"'
                f' domain="{replica["domain"]}"'
                f' priority="{replica["priority"]}"'
                f' client_extract="{str(replica["client_extract"]).lower()}"'
                f'>{escape(pfn)}</url>\n'
            )
        else:
            yield (
                '  '
                f'<url location="{replica["rse"]}"'
                f' priority="{replica["priority"]}"'
                f'>{escape(pfn)}</url>\n'
            )
    yield ' </file>\n'


def _generate_metalink_response(rfiles, policy_schema, detailed_url=True):
    first = True
    for rfile in rfiles:
        if first:
            # first, set the appropriate content type, and stream the header
            yield '<?xml version="1.0" encoding="UTF-8"?>\n<metalink xmlns="urn:ietf:params:xml:ns:metalink">\n'
            first = False

        yield from _generate_one_metalink_file(rfile, policy_schema=policy_schema, detailed_url=detailed_url)

    if first:
        # if still first output, i.e. there were no replicas
        yield '<?xml version="1.0" encoding="UTF-8"?>\n<metalink xmlns="urn:ietf:params:xml:ns:metalink">\n</metalink>\n'
    else:
        # don't forget to send the metalink footer
        yield '</metalink>\n'


def _generate_json_response(rfiles):
    for rfile in rfiles:
        yield dumps(rfile) + '\n'


class Replicas(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream', 'application/metalink4+xml'])
    def get(self, scope_name):
        """
        ---
        summary: Get Replicas
        description: List all replicas for data identifiers.
        tags:
          - Replicas
        parameters:
        - name: scope_name
          in: path
          description: The DID associated with the replicas.
          schema:
            type: string
          style: simple
        - name: X-Forwarded-For
          in: header
          description: The client ip
          schema:
            type: string
        - name: schemes
          in: query
          description: The schemes of the replicas.
          schema:
            type: string
        - name: select
          in: query
          description: The sorting algorithm.
          schema:
            type: string
            enum: ["geoip", "closeness", "dynamic", "ranking", "random"]
        - name: limit
          in: query
          description: The maximum number of replicas returned.
          schema:
            type: integer
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list with all replicas.
                  type: array
                  items:
                    description: A replica. Possibly contains more information.
                    type: object
                    properties:
                      scope:
                        description: The scope of the replica.
                        type: string
                      name:
                        description: The name of the replica.
                        type: string
                      bytes:
                        description: The size of the replica in bytes.
                        type: integer
                      md5:
                        description: The md5 checksum of the replica.
                        type: string
                      adler32:
                        description: The adler32 checksum of the replica.
                        type: string
                      pfns:
                        description: The pfns associated with the replica.
                        type: array
                      rses:
                        description: The rse associated with the replica.
                        type: string
          401:
            description: Invalid Auth Token
          404:
            description: Did not found
          406:
            description: Not acceptable
        """
        try:
            scope, name = parse_scope_name(scope_name, request.environ.get('vo'))
        except ValueError as error:
            return generate_http_error_flask(400, error)

        content_type = request.accept_mimetypes.best_match(['application/x-json-stream', 'application/metalink4+xml'], 'application/x-json-stream')
        metalink = (content_type == 'application/metalink4+xml')
        dids = [{'scope': scope, 'name': name}]
        select, limit = None, None

        client_ip = request.headers.get('X-Forwarded-For', default=request.remote_addr)
        client_location = {'ip': client_ip, 'fqdn': None, 'site': None}

        schemes = request.args.get('schemes', default=None)
        select = request.args.get('select', default=None)
        limit = request.args.get('limit', default=None)
        if limit:
            limit = int(limit)

        # Resolve all reasonable protocols when doing metalink for maximum access possibilities
        if metalink and schemes is None:
            schemes = SUPPORTED_PROTOCOLS

        try:
            def _list_and_sort_replicas(vo):
                # we need to call list_replicas before starting to reply
                # otherwise the exceptions won't be propagated correctly
                for rfile in list_replicas(dids=dids, schemes=schemes, vo=vo):
                    replicas = []
                    dictreplica = {}
                    for rse in rfile['rses']:
                        for replica in rfile['rses'][rse]:
                            replicas.append(replica)
                            dictreplica[replica] = rse

                    replicas = sort_replicas(dictreplica, client_location, selection=select)
                    rfile['pfns'] = dict(_sorted_with_priorities(rfile['pfns'], replicas, limit=limit))
                    yield rfile

            rfiles = _list_and_sort_replicas(vo=request.environ.get('vo'))
            if metalink:
                response_generator = _generate_metalink_response(rfiles, 'atlas', detailed_url=False)
            else:
                response_generator = _generate_json_response(rfiles)
            return try_stream(response_generator, content_type=content_type)
        except DataIdentifierNotFound as error:
            return generate_http_error_flask(404, error)

    def post(self):
        """
        ---
        summary: Create File Replicas
        description: Create file replicas at a given RSE.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                - rse
                - files
                properties:
                  rse:
                    description: The rse for the replication
                    type: string
                  files:
                    description: The files to replicate
                    type: array
                    items:
                      type: object
                      required:
                        - pfn
                        - bytes
                        - name
                      properties:
                        pfn:
                          description: The pfn of the replica.
                          type: string
                        name:
                          description: The DID name.
                          type: string
                        bytes:
                          description: The size of the replica in bytes.
                          type: integer
                        state:
                          description: The state of the replica.
                          type: string
                        path:
                          description: The path of the new replica.
                          type: string
                        md5:
                          description: The md5 checksum.
                          type: string
                        adler32:
                          description: The adler32 checksum.
                          type: string
                        lcok_cnt:
                          description: The lock count.
                          type: integer
                        tombstone:
                          description: The tombstone.
                          type: string
                  ignore_availability:
                    description: The ignore availability.
                    type: boolean
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ["Created"]
          400:
            description: Invalid Path
          401:
            description: Invalid Auth Token
          404:
            description: Rse or scope not found
          409:
            description: Replica or Did already exists
          503:
            description: Resource temporary unavailable
        """
        parameters = json_parameters(parse_response)
        rse = param_get(parameters, 'rse')
        files = param_get(parameters, 'files')

        try:
            add_replicas(
                rse=rse,
                files=files,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo'),
                ignore_availability=param_get(parameters, 'ignore_availability', default=False),
            )
        except InvalidPath as error:
            return generate_http_error_flask(400, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except (Duplicate, DataIdentifierAlreadyExists) as error:
            return generate_http_error_flask(409, error)
        except (RSENotFound, ScopeNotFound) as error:
            return generate_http_error_flask(404, error)
        except ResourceTemporaryUnavailable as error:
            return generate_http_error_flask(503, error)

        return 'Created', 201

    def put(self):
        """
        ---
        summary: Update File Replicas
        description: Update file replicas state at a given RSE.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                - rse
                - files
                properties:
                  rse:
                    description: The rse for the replication
                    type: string
                  files:
                    description: The files to replicate
                    type: array
                    items:
                      type: object
                      properties:
                        name:
                          description: The pfn of the replica.
                          type: string
                        state:
                          description: The pfn of the replica.
                          type: string
                        path:
                          description: The pfn of the replica.
                          type: string
                        error_message:
                          description: The error message if an error occured.
                          type: string
                        broken_rule_id:
                          description: The id of the broken rule if one was found.
                          type: string
                        broken_message:
                          description: The message of the broken rule.
                          type: string
        responses:
          200:
            description: OK
          400:
            description: Cannot decode json parameter list
          401:
            description: Invalid Auth Token
        """
        parameters = json_parameters(parse_response)
        rse = param_get(parameters, 'rse')
        files = param_get(parameters, 'files')

        try:
            update_replicas_states(rse=rse, files=files, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)

        return '', 200

    def delete(self):
        """
        ---
        summary: Delete File Replicas
        description: Delete file replicas at a given RSE.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                  - rse
                  - files
                properties:
                  rse:
                    description: The rse name.
                    type: string
                  files:
                    description: The files to delete.
                    type: array
                    items:
                      type: object
                      required:
                        - name
                      properties:
                        name:
                          description: The name of the replica.
                          type: string
        responses:
          200:
            description: OK
          400:
            description: Cannot decode json parameter list.
          401:
            description: Invalid Auth Token
          404:
            description: Rse or Replica not found
        """
        parameters = json_parameters(parse_response)
        rse = param_get(parameters, 'rse')
        files = param_get(parameters, 'files')

        try:
            delete_replicas(
                rse=rse,
                files=files,
                issuer=request.environ.get('issuer'),
                vo=request.environ.get('vo'),
                ignore_availability=param_get(parameters, 'ignore_availability', default=False),
            )
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except (RSENotFound, ReplicaNotFound) as error:
            return generate_http_error_flask(404, error)
        except ResourceTemporaryUnavailable as error:
            return generate_http_error_flask(503, error)

        return '', 200


class ListReplicas(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream', 'application/metalink4+xml'])
    def post(self):
        """
        ---
        summary: List Replicas
        description: List all replicas for a DID.
        tags:
          - Replicas
        parameters:
        - name: X-Forwarded-For
          in: header
          description: The client ip address.
          schema:
            type: string
        - name: limit
          in: query
          description: The maximum number pfns per replica to return.
          schema:
            type: integer
        - name: select
          in: query
          description: Requested sorting of the result, e.g., 'geoip', 'closeness', 'dynamic', 'ranking', 'random'.
          schema:
            type: string
        - name: sort
          in: query
          description: Requested sorting of the result, e.g., 'geoip', 'closeness', 'dynamic', 'ranking', 'random'.
          schema:
            type: string
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  client_location:
                    description: The clients location.
                    type: string
                  dids:
                    description: List of Dids.
                    type: array
                    items:
                      type: object
                      properties:
                        scope:
                          description: The scope of the did.
                          type: string
                        name:
                          description: The name of the did.
                          type: string
                  schemes:
                    description: A list of schemes to filter the replicas.
                    type: array
                    items:
                      type: string
                  sort:
                    description: Requested sorting of the result, e.g., 'geoip', 'closeness', 'dynamic', 'ranking', 'random'.
                    type: string
                  unavailable:
                    description: If unavailable rse should be considered.
                    type: boolean
                    deprecated: true
                  ignore_availability:
                    description: If the availability should be ignored.
                    type: boolean
                  rse_expression:
                    description: The RSE expression to restrict on a list of RSEs.
                    type: string
                  all_states:
                    description: Return all replicas whatever state they are in. Adds an extra 'states' entry in the result dictionary.
                    type: boolean
                  domain:
                    description: The network domain for the call, either None, 'wan' or 'lan'. None is fallback to 'wan', 'all' is both ['lan','wan']
                    type: string
                  signature_lifetime:
                    description: If supported, in seconds, restrict the lifetime of the signed PFN.
                    type: integer
                  resolve_archives:
                    description:  When set to True, find archives which contain the replicas.
                    type: boolean
                  resolve_parents:
                    description: When set to True, find all parent datasets which contain the replicas.
                    type: boolean
                  updated_after:
                    description: datetime object (UTC time), only return replicas updated after this time
                    type: string
                  nrandom:
                    description: The maximum number of replicas to return.
                    type: integer
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      scope:
                        description: The scope of the replica.
                        type: string
                      name:
                        description: The name of the replica.
                        type: string
                      bytes:
                        description: The size of the replica in bytes.
                        type: integer
                      md5:
                        description: The md5 checksum.
                        type: string
                      adler32:
                        description: The adler32 checksum.
                        type: string
                      pfns:
                        description: The pfns.
                        type: array
                      rses:
                        description: The RSESs.
                        type: array
              application/metalink4+xml:
                schema:
                  type: object
                  properties:
                    scope:
                      description: The scope of the replica.
                      type: string
                    name:
                      description: The name of the replica.
                      type: string
                    bytes:
                      description: The size of the replica in bytes.
                      type: integer
                    md5:
                      description: The md5 checksum.
                      type: string
                    adler32:
                      description: The adler32 checksum.
                      type: string
                    pfns:
                      description: The pfns.
                      type: array
                    rses:
                      description: The RSESs.
                      type: array
          400:
            description: Cannot decode json parameter list.
          401:
            description: Invalid Auth Token
          404:
            description: Did not found.
          406:
            description: Not acceptable
        """
        content_type = request.accept_mimetypes.best_match(['application/x-json-stream', 'application/metalink4+xml'], 'application/x-json-stream')
        metalink = (content_type == 'application/metalink4+xml')

        client_ip = request.headers.get('X-Forwarded-For', default=request.remote_addr)

        parameters = json_parameters(parse_response)

        client_location = {'ip': client_ip,
                           'fqdn': None,
                           'site': None}
        client_location.update(param_get(parameters, 'client_location', default={}))

        # making sure IP address is not overwritten
        client_location['ip'] = client_ip

        dids = param_get(parameters, 'dids', default=[])
        schemes = param_get(parameters, 'schemes', default=None)
        select = param_get(parameters, 'sort', default=None)
        unavailable = param_get(parameters, 'unavailable', default=False)
        ignore_availability = param_get(parameters, 'ignore_availability', default='unavailable' in parameters)
        rse_expression = param_get(parameters, 'rse_expression', default=None)
        all_states = param_get(parameters, 'all_states', default=False)
        domain = param_get(parameters, 'domain', default=None)
        if 'signature_lifetime' in parameters:
            signature_lifetime = param_get(parameters, 'signature_lifetime')
        else:
            # hardcoded default of 10 minutes if config is not parseable
            signature_lifetime = config_get_int('credentials', 'signature_lifetime', raise_exception=False, default=600)
        resolve_archives = param_get(parameters, 'resolve_archives', default=True)
        resolve_parents = param_get(parameters, 'resolve_parents', default=False)
        updated_after = param_get(parameters, 'updated_after', default=None)
        if updated_after is not None:
            if isinstance(updated_after, (int, float)):
                # convert from epoch time stamp to datetime object
                updated_after = datetime.utcfromtimestamp(updated_after)
            else:
                # attempt UTC format '%Y-%m-%dT%H:%M:%S' conversion
                updated_after = datetime.strptime(updated_after, '%Y-%m-%dT%H:%M:%S')
        nrandom = param_get(parameters, 'nrandom', default=None)
        if nrandom:
            nrandom = int(nrandom)

        limit = request.args.get('limit', default=None)
        select = request.args.get('select', default=select)
        select = request.args.get('sort', default=select)

        # Resolve all reasonable protocols when doing metalink for maximum access possibilities
        if metalink and schemes is None:
            schemes = SUPPORTED_PROTOCOLS

        content_type = 'application/metalink4+xml' if metalink else 'application/x-json-stream'

        try:
            def _list_and_sort_replicas(request_id, issuer, vo):
                # we need to call list_replicas before starting to reply
                # otherwise the exceptions won't be propagated correctly
                for rfile in list_replicas(dids=dids, schemes=schemes,
                                           unavailable=unavailable,
                                           request_id=request_id,
                                           ignore_availability=ignore_availability,
                                           all_states=all_states,
                                           rse_expression=rse_expression,
                                           client_location=client_location,
                                           domain=domain,
                                           signature_lifetime=signature_lifetime,
                                           resolve_archives=resolve_archives,
                                           resolve_parents=resolve_parents,
                                           nrandom=nrandom,
                                           updated_after=updated_after,
                                           issuer=issuer,
                                           vo=vo):

                    # Sort rfile['pfns'] and limit its size according to "limit" parameter
                    lanreplicas = {}
                    wanreplicas = {}
                    for pfn, replica in rfile['pfns'].items():
                        replica_tuple = (replica['domain'], replica['priority'], replica['rse'], replica['client_extract'])
                        if replica_tuple[0] == 'lan':
                            lanreplicas[pfn] = replica_tuple
                        else:
                            wanreplicas[pfn] = replica_tuple

                    rfile['pfns'] = dict(_sorted_with_priorities(replicas=rfile['pfns'],
                                                                 # Lan replicas sorted by priority; followed by wan replicas sorted by selection criteria
                                                                 sorted_pfns=chain(sorted(lanreplicas.keys(), key=lambda pfn: lanreplicas[pfn][1]),
                                                                                   sort_replicas(wanreplicas, client_location, selection=select)),
                                                                 limit=limit))
                    yield rfile

            rfiles = _list_and_sort_replicas(request_id=request.environ.get('request_id'),
                                             issuer=request.environ.get('issuer'),
                                             vo=request.environ.get('vo'))
            if metalink:
                policy_schema = config_get('policy', 'schema', raise_exception=False, default='generic')
                response_generator = _generate_metalink_response(rfiles, policy_schema)
            else:
                response_generator = _generate_json_response(rfiles)
            return try_stream(response_generator, content_type=content_type)
        except InvalidObject as error:
            return generate_http_error_flask(400, error)
        except DataIdentifierNotFound as error:
            return generate_http_error_flask(404, error)


class ReplicasDIDs(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def post(self):
        """
        ---
        summary: List Replicas Dids
        description: List the DIDs associated to a list of replicas.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                - rse
                properties:
                  pfns:
                    description: The list of pfns.
                    type: array
                    items:
                      type: string
                  rse:
                    description: The RSE name.
                    type: string
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  type: array
                  items:
                    type: object
                    additionalProperties:
                      x-additionalPropertiesName: mapped PFNs to DIDs
                      description: A mapping from a pfn to a did.
                      type: object
                      properties:
                        scope:
                          description: The scope of the DID.
                          type: string
                        name:
                          description: The name of the DID.
                          type: string
          400:
            description: Cannot decode json parameter list.
          401:
            description: Invalid Auth Token
          404:
            description: Not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()
        pfns = param_get(parameters, 'pfns', default=[])
        rse = param_get(parameters, 'rse')

        try:
            def generate(vo):
                for pfn in get_did_from_pfns(pfns, rse, vo=vo):
                    yield dumps(pfn) + '\n'

            return try_stream(generate(vo=request.environ.get('vo')))
        except AccessDenied as error:
            return generate_http_error_flask(401, error)


class BadReplicas(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/json'])
    def post(self):
        """
        ---
        summary: Declare Bad Replicas
        description: Declares a list of bad replicas.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  replicas:
                    description: The list of pfns or list of dicts with "scope", "name", "rse_id"/"rse"
                    type: array
                    items:
                      type: string
                  pfns:
                    deprecated: true
                    description: The list of pfns, for backward compatibility with older versions of the ReplicaClient
                    type: array
                    items:
                      type: string
                  reason:
                    description: The reason for the declaration.
                    type: string
                  force:
                    description: If true, ignore existing replica status in the bad_replicas table.
                    type: boolean
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  description: Returns the not declared files.
                  type: array
          400:
            description: Can not decode json parameter list.
          404:
            description: Not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters()
        replicas = param_get(parameters, 'replicas', default=[]) or param_get(parameters, 'pfns', default=[])
        reason = param_get(parameters, 'reason', default=None)
        force = param_get(parameters, 'force', default=False)

        try:
            not_declared_files = declare_bad_file_replicas(replicas, reason=reason,
                                                           issuer=request.environ.get('issuer'), vo=request.environ.get('vo'),
                                                           force=force)
            return not_declared_files, 201
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except (RSENotFound, ReplicaNotFound) as error:
            return generate_http_error_flask(404, error)


class QuarantineReplicas(ErrorHandlingMethodView):

    def post(self):
        """
        ---
        summary: Quarantine replicas
        description: Quarantine replicas.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                    - replicas
                properties:
                  replicas:
                    description: replicas
                    type: array
                    items:
                      type: object
                      required:
                        - path
                      properties:
                            path:
                                description: path
                                type:   string
                            scope:
                                description: scope
                                type:   string
                            name:
                                description: name
                                type:   string
                  rse:
                    description: RSE name
                    type: string
                  rse_id:
                    description: RSE id
                    type: string
        responses:
          200:
            description: OK
          403:
            description: Forbidden.
          404:
            description: Not found
        """

        parameters = json_parameters()
        replicas = param_get(parameters, 'replicas', default=[])
        rse = param_get(parameters, 'rse', default=None)
        rse_id = param_get(parameters, 'rse_id', default=None)
        vo = request.environ.get('vo')
        issuer = request.environ.get('issuer')

        if replicas:
            try:
                quarantine_file_replicas(replicas, issuer, rse=rse, rse_id=rse_id, vo=vo)
            except AccessDenied as error:
                return generate_http_error_flask(403, error)
            except (RSENotFound, ReplicaNotFound) as error:
                return generate_http_error_flask(404, error)

        return '', 200


class SuspiciousReplicas(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/json'])
    def post(self):
        """
        ---
        summary: Declare Suspicious Replicas
        description: Declare a list of suspicious replicas.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  pfns:
                    description: The list of pfns.
                    type: array
                    items:
                      type: string
                  reason:
                    description: The reason for the declaration.
                    type: string
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  description: Returns the not declared files.
                  type: array
          400:
            description: Can not decode json parameter list.
          404:
            description: Not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters(parse_response)
        pfns = param_get(parameters, 'pfns', default=[])
        reason = param_get(parameters, 'reason', default=None)

        try:
            not_declared_files = declare_suspicious_file_replicas(pfns=pfns, reason=reason, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
            return not_declared_files, 201
        except AccessDenied as error:
            return generate_http_error_flask(401, error)

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self):
        """
        ---
        summary: List Suspicious Replicas
        description: List the suspicious replicas on a list of RSEs.
        tags:
          - Replicas
        parameters:
        - name: rse_expression
          in: query
          description: The RSE expression to filter for.
          schema:
            type: string
        - name: younger_than
          in: query
          description: Date to filter for.
          schema:
            type: string
        - name: nattempts
          in: query
          description: The maximum number of attempts to make.
          schema:
            type: integer
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      scope:
                        description: The scope of the Replica.
                        type: string
                      name:
                        description: The name of the Replica.
                        type: string
                      rse:
                        description: The rse name.
                        type: string
                      rse_id:
                        description: The id of the rse.
                        type: string
                      cnt:
                        description: The number of replicas.
                        type: integer
                      created_at:
                        description: The time when the replica was created.
                        type: string
          401:
            description: Invalid Auth Token
          404:
            description: Not found
          406:
            description: Not acceptable
        """
        rse_expression, younger_than, nattempts = None, None, None
        if request.query_string:
            query_string = request.query_string.decode(encoding='utf-8')
            try:
                params = loads(unquote(query_string))
            except ValueError:
                params = parse_qs(query_string)

            if 'rse_expression' in params:
                rse_expression = params['rse_expression'][0]
            if 'younger_than' in params and params['younger_than'][0]:
                younger_than = datetime.strptime(params['younger_than'][0], "%Y-%m-%dT%H:%M:%S")
            if 'nattempts' in params:
                nattempts = int(params['nattempts'][0])

        result = get_suspicious_files(rse_expression=rse_expression, younger_than=younger_than, nattempts=nattempts, vo=request.environ.get('vo'))
        return Response(render_json_list(result), 200, content_type='application/json')


class BadReplicasStates(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: List Bad Replicas By States
        description: List the bad or suspicious replicas by states.
        tags:
          - Replicas
        parameters:
        - name: state
          in: query
          description: The state of the file.
          schema:
            type: string
            enum: [SUSPICIOUS, BAD]
        - name: rse
          in: query
          description: The rse name.
          schema:
            type: string
        - name: younger_than
          in: query
          description: Date to select bad replicas younger than this date.
          schema:
            type: string
            format: date-time
        - name: older_than
          in: query
          description: Date to select bad replicas older than this date.
          schema:
            type: string
            format: date-time
        - name: limit
          in: query
          description: The maximum number of replicas returned.
          schema:
            type: integer
        - name: list_pfns
          in: query
          description: Flag to include pfns.
          schema:
            type: boolean
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list of all result replicas.
                  type: array
                  items:
                    oneOf:
                      - type: object
                        properties:
                          scope:
                            description: The scope fo the replica.
                            type: string
                          name:
                            description: The name of the replica.
                            type: string
                          type:
                            description: The type of the replica.
                            type: string
                      - type: object
                        properties:
                          scope:
                            description: The scope fo the replica.
                            type: string
                          name:
                            description: The name of the replica.
                            type: string
                          rse:
                            description: The name of the associated rse.
                            type: string
                          rse_id:
                            description: The id of the associated rse.
                            type: string
                          state:
                            description: The state of the replica.
                            type: string
                          created_at:
                            description: The date-time the replica was created.
                            type: string
                            format: date-time
                          updated_at:
                            description: The date-time the replica was updated.
                            type: string
                            format: date-time
          401:
            description: Invalid Auth Token
          406:
            description: Not acceptable
        """
        state, rse, younger_than, older_than, limit, list_pfns = None, None, None, None, None, None
        if request.query_string:
            query_string = request.query_string.decode(encoding='utf-8')
            try:
                params = loads(unquote(query_string))
            except ValueError:
                params = parse_qs(query_string)
            if 'state' in params:
                state = params['state'][0]
            if isinstance(state, str):
                state = BadFilesStatus(state)
            if 'rse' in params:
                rse = params['rse'][0]
            if 'younger_than' in params:
                younger_than = datetime.strptime(params['younger_than'][0], "%Y-%m-%dT%H:%M:%S.%f")
            if 'older_than' in params and params['older_than']:
                older_than = datetime.strptime(params['older_than'][0], "%Y-%m-%dT%H:%M:%S.%f")
            if 'limit' in params:
                limit = int(params['limit'][0])
            if 'list_pfns' in params:
                list_pfns = bool(params['list_pfns'][0])

        def generate(vo):
            for row in list_bad_replicas_status(state=state, rse=rse, younger_than=younger_than,
                                                older_than=older_than, limit=limit, list_pfns=list_pfns,
                                                vo=vo):
                yield dumps(row, cls=APIEncoder) + '\n'

        return try_stream(generate(vo=request.environ.get('vo')))


class BadReplicasSummary(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: Bad Replicas Summary
        description: Return a summary of the bad replicas by incident.
        tags:
          - Replicas
        parameters:
        - name: rse_expression
          in: query
          description: The RSE expression.
          schema:
            type: string
        - name: from_date
          in: query
          description: The start date.
          schema:
            type: string
            format: date-time
        - name: to_date
          in: query
          description: The end date.
          schema:
            type: string
            format: date-time
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list of summaries.
                  type: array
                  items:
                    type: object
                    properties:
                      rse:
                        description: The name of the associated RSE.
                        type: string
                      rse_id:
                        description: The id of the associated RSE.
                        type: string
                      created_at:
                        description: The creation date-time.
                        type: string
                        format: date-time
                      reason:
                        description: The reason for the incident.
                        type: string
          401:
            description: Invalid Auth Token
          406:
            description: Not acceptable
        """
        rse_expression, from_date, to_date = None, None, None
        if request.query_string:
            query_string = request.query_string.decode(encoding='utf-8')
            try:
                params = loads(unquote(query_string))
            except ValueError:
                params = parse_qs(query_string)
            if 'rse_expression' in params:
                rse_expression = params['rse_expression'][0]
            if 'from_date' in params and params['from_date'][0]:
                from_date = datetime.strptime(params['from_date'][0], "%Y-%m-%d")
            if 'to_date' in params:
                to_date = datetime.strptime(params['to_date'][0], "%Y-%m-%d")

        def generate(vo):
            for row in get_bad_replicas_summary(rse_expression=rse_expression, from_date=from_date,
                                                to_date=to_date, vo=vo):
                yield dumps(row, cls=APIEncoder) + '\n'

        return try_stream(generate(vo=request.environ.get('vo')))


class DatasetReplicas(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, scope_name):
        """
        ---
        summary: List Dataset Replicas
        description: List dataset replicas.
        tags:
          - Replicas
        parameters:
        - name: scope_name
          in: path
          description: data identifier (scope)/(name).
          schema:
            type: string
          style: simple
        - name: deep
          in: query
          description: Flag to ennable lookup at the file level.
          schema:
            type: boolean
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list of dataset replicas.
                  type: array
                  items:
                    type: object
                    properties:
                      scope:
                        description: The scope of the replica.
                        type: string
                      name:
                        description: The name of the replica.
                        type: string
                      rse:
                        description: The name of the associated RSE.
                        type: string
                      rse_id:
                        description: The id of the associated RSE.
                        type: string
                      bytes:
                        description: The size of the replica.
                        type: integer
                      length:
                        description: The length of the replica.
                        type: integer
                      available_bytes:
                        description: The number of available bytes of the replica.
                        type: integer
                      available_length:
                        description: The available length of the replica.
                        type: integer
                      state:
                        description: The state of the replica.
                        type: string
                      created_at:
                        description: The date-time the replica was created.
                        type: string
                        format: date-time
                      updated_at:
                        description: The date-time the replica was updated.
                        type: string
                        format: date-time
                      accessed_at:
                        description: The date-time the replica was accessed.
                        type: string
                        format: date-time
          401:
            description: Invalid Auth Token
          404:
            description: Not found
          406:
            description: Not acceptable
        """
        try:
            scope, name = parse_scope_name(scope_name, request.environ.get('vo'))

            def generate(_deep, vo):
                for row in list_dataset_replicas(scope=scope, name=name, deep=_deep, vo=vo):
                    yield dumps(row, cls=APIEncoder) + '\n'

            deep = request.args.get('deep', default=False)

            return try_stream(generate(_deep=deep, vo=request.environ.get('vo')))
        except ValueError as error:
            return generate_http_error_flask(400, error)


class DatasetReplicasBulk(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def post(self):
        """
        ---
        summary: List Dataset Replicas for Multiple DIDs
        description: List dataset replicas for multiple dids.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                required:
                - dids
                properties:
                  dids:
                    description: A list of dids.
                    type: array
                    items:
                      description: A did.
                      type: object
                      properties:
                        scope:
                          description: The scope of the did.
                          type: string
                        name:
                          description: The name of the did.
                          type: string
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list of dataset replicas.
                  type: array
                  items:
                    type: object
                    properties:
                      scope:
                        description: The scope of the replica.
                        type: string
                      name:
                        description: The name of the replica.
                        type: string
                      rse:
                        description: The name of the associated RSE.
                        type: string
                      rse_id:
                        description: The id of the associated RSE.
                        type: string
                      bytes:
                        description: The size of the replica.
                        type: integer
                      length:
                        description: The length of the replica.
                        type: integer
                      available_bytes:
                        description: The number of available bytes of the replica.
                        type: integer
                      available_length:
                        description: The available length of the replica.
                        type: integer
                      state:
                        description: The state of the replica.
                        type: string
                      created_at:
                        description: The date-time the replica was created.
                        type: string
                        format: date-time
                      updated_at:
                        description: The date-time the replica was updated.
                        type: string
                        format: date-time
                      accessed_at:
                        description: The date-time the replica was accessed.
                        type: string
                        format: date-time
          400:
            description: Bad Request.
          401:
            description: Invalid Auth Token
          404:
            description: Not found
          406:
            description: Not acceptable
        """
        parameters = json_parameters(parse_response)
        dids = param_get(parameters, 'dids')
        if len(dids) == 0:
            return generate_http_error_flask(400, ValueError.__name__, 'List of DIDs is empty')

        try:
            def generate(vo):
                for row in list_dataset_replicas_bulk(dids=dids, vo=vo):
                    yield dumps(row, cls=APIEncoder) + '\n'

            return try_stream(generate(vo=request.environ.get('vo')))
        except InvalidObject as error:
            return generate_http_error_flask(400, error, f'Cannot validate DIDs: {error}')


class DatasetReplicasVP(ErrorHandlingMethodView):
    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, scope_name):
        """
        ---
        summary: List Dataset Replicas VP
        description: |
          List dataset replicas using the Virtual Placement service.
          This is an RnD function and might change or go away at any time.
        tags:
          - Replicas
        parameters:
        - name: scope_name
          in: path
          description: data identifier (scope)/(name).
          schema:
            type: string
          style: simple
        - name: deep
          in: query
          description: Flag to ennable lookup at the file level.
          schema:
            type: boolean
        responses:
          200:
            description: OK. This needs documentation!
          401:
            description: Invalid Auth Token
          406:
            description: Not acceptable
        """
        try:
            scope, name = parse_scope_name(scope_name, request.environ.get('vo'))

            def generate(_deep, vo):
                for row in list_dataset_replicas_vp(scope=scope, name=name, deep=_deep, vo=vo):
                    yield dumps(row, cls=APIEncoder) + '\n'

            deep = request.args.get('deep', default=False)

            return try_stream(generate(_deep=deep, vo=request.environ.get('vo')))
        except ValueError as error:
            return generate_http_error_flask(400, error)


class ReplicasRSE(ErrorHandlingMethodView):

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, rse):
        """
        ---
        summary: List Dataset Replicas per RSE
        description: List dataset replicas per RSE.
        tags:
          - Replicas
        parameters:
        - name: rse
          in: path
          description: The rse to filter for.
          schema:
            type: string
          style: simple
        responses:
          200:
            description: OK
            content:
              application/x-json-stream:
                schema:
                  description: A list of dataset replicas.
                  type: array
                  items:
                    type: object
                    properties:
                      scope:
                        description: The scope of the replica.
                        type: string
                      name:
                        description: The name of the replica.
                        type: string
                      rse:
                        description: The name of the associated RSE.
                        type: string
                      rse_id:
                        description: The id of the associated RSE.
                        type: string
                      bytes:
                        description: The size of the replica.
                        type: integer
                      length:
                        description: The length of the replica.
                        type: integer
                      available_bytes:
                        description: The number of available bytes of the replica.
                        type: integer
                      available_length:
                        description: The available length of the replica.
                        type: integer
                      state:
                        description: The state of the replica.
                        type: string
                      created_at:
                        description: The date-time the replica was created.
                        type: string
                        format: date-time
                      updated_at:
                        description: The date-time the replica was updated.
                        type: string
                        format: date-time
                      accessed_at:
                        description: The date-time the replica was accessed.
                        type: string
                        format: date-time
          401:
            description: Invalid Auth Token
          406:
            description: Not acceptable
        """

        def generate(vo):
            for row in list_datasets_per_rse(rse=rse, vo=vo):
                yield dumps(row, cls=APIEncoder) + '\n'

        return try_stream(generate(vo=request.environ.get('vo')))


class BadDIDs(ErrorHandlingMethodView):

    def post(self):
        """
        ---
        summary: Mark Bad by DID
        description: Declare a list of bad replicas by DID.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  expires_at:
                    description: The expires at value.
                    type: string
                    format: date-time
                  dids:
                    description: The list of dids associated with the bad replicas.
                    type: array
                    items:
                      type: object
                      properties:
                        scope:
                          description: The scope of the did.
                          type: string
                        name:
                          description: The name of the did.
                          type: string
                  rse:
                    description: The name of the rse.
                    type: string
                  reason:
                    description: The reason for the change.
                    type: string
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  description: All files not declared as bad.
                  type: array
                  items:
                    type: string
          400:
            description: Cannot decode json parameter list.
          401:
            description: Invalid Auth Token
          404:
            description: Not found
        """
        parameters = json_parameters(parse_response)
        expires_at = param_get(parameters, 'expires_at', default=None)
        if expires_at:
            expires_at = datetime.strptime(expires_at, "%Y-%m-%dT%H:%M:%S.%f")

        try:
            not_declared_files = add_bad_dids(
                dids=param_get(parameters, 'dids', default=[]),
                rse=param_get(parameters, 'rse', default=None),
                issuer=request.environ.get('issuer'),
                state=BadFilesStatus.BAD,
                reason=param_get(parameters, 'reason', default=None),
                expires_at=expires_at,
                vo=request.environ.get('vo'),
            )
        except (ValueError, InvalidType) as error:
            return generate_http_error_flask(400, ValueError.__name__, error.args[0])
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except ReplicaNotFound as error:
            return generate_http_error_flask(404, error)
        except Duplicate as error:
            return generate_http_error_flask(409, error)

        return Response(dumps(not_declared_files), status=201, content_type='application/json')


class BadPFNs(ErrorHandlingMethodView):

    def post(self):
        """
        ---
        summary: Declare Bad PFNs
        description: Declare a list of bad PFNs.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  expires_at:
                    description: The expires at value. Only apply to TEMPORARY_UNAVAILABLE.
                    type: string
                    format: date-time
                  pfns:
                    description: The list of pfns associated with the bad PFNs.
                    type: array
                    items:
                      type: string
                  state:
                    description: The state to set the PFNs to.
                    type: string
                    enum: ["BAD", "SUSPICIOUS", "TEMPORARY_UNAVAILABLE"]
                  reason:
                    description: The reason for the change.
                    type: string
        responses:
          201:
            description: Created
          400:
            description: Cannot decode json parameter list.
          401:
            description: Invalid Auth Token
          404:
            description: Replica not found
          409:
            description: Duplicate
        """
        parameters = json_parameters(parse_response)
        expires_at = param_get(parameters, 'expires_at', default=None)
        if expires_at:
            expires_at = datetime.strptime(expires_at, "%Y-%m-%dT%H:%M:%S.%f")

        try:
            add_bad_pfns(
                pfns=param_get(parameters, 'pfns', default=[]),
                issuer=request.environ.get('issuer'),
                state=param_get(parameters, 'state', default=None),
                reason=param_get(parameters, 'reason', default=None),
                expires_at=expires_at,
                vo=request.environ.get('vo'),
            )
        except (ValueError, InvalidType) as error:
            return generate_http_error_flask(400, ValueError.__name__, error.args[0])
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except ReplicaNotFound as error:
            return generate_http_error_flask(404, error)
        except Duplicate as error:
            return generate_http_error_flask(409, error)

        return 'Created', 201


class Tombstone(ErrorHandlingMethodView):

    def post(self):
        """
        ---
        summary: Set Tombstone
        description: Set a tombstone on a list of replicas.
        tags:
          - Replicas
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  replicas:
                    description: The replicas to set the tombstone to.
                    type: array
                    items:
                      type: object
                      required:
                        - rse
                        - scope
                        - name
                      properties:
                        rse:
                          description: The rse associated with the tombstone.
                          type: string
                        scope:
                          description: The scope of the replica
                          type: string
                        name:
                          description: The name of the replica.
                          type: string
        responses:
          201:
            description: Created
          401:
            description: Invalid Auth Token
          404:
            description: Not found
          423:
            description: Replica is locked.
        """
        parameters = json_parameters(parse_response)
        replicas = param_get(parameters, 'replicas', default=[])

        try:
            for replica in replicas:
                set_tombstone(
                    rse=replica['rse'],
                    scope=replica['scope'],
                    name=replica['name'],
                    issuer=request.environ.get('issuer'),
                    vo=request.environ.get('vo'),
                )
        except ReplicaNotFound as error:
            return generate_http_error_flask(404, error)
        except ReplicaIsLocked as error:
            return generate_http_error_flask(423, error)

        return 'Created', 201


def blueprint(with_doc=False):
    bp = AuthenticatedBlueprint('replicas', __name__, url_prefix='/replicas')

    list_replicas_view = ListReplicas.as_view('list_replicas')
    bp.add_url_rule('/list', view_func=list_replicas_view, methods=['post', ])
    replicas_view = Replicas.as_view('replicas')
    if not with_doc:
        # rule without trailing slash needs to be added before rule with trailing slash
        bp.add_url_rule('', view_func=replicas_view, methods=['post', 'put', 'delete'])
    bp.add_url_rule('/', view_func=replicas_view, methods=['post', 'put', 'delete'])
    suspicious_replicas_view = SuspiciousReplicas.as_view('suspicious_replicas')
    bp.add_url_rule('/suspicious', view_func=suspicious_replicas_view, methods=['get', 'post'])
    bad_replicas_states_view = BadReplicasStates.as_view('bad_replicas_states')
    bp.add_url_rule('/bad/states', view_func=bad_replicas_states_view, methods=['get', ])
    bad_replicas_summary_view = BadReplicasSummary.as_view('bad_replicas_summary')
    bp.add_url_rule('/bad/summary', view_func=bad_replicas_summary_view, methods=['get', ])
    bad_replicas_pfn_view = BadPFNs.as_view('add_bad_pfns')
    bp.add_url_rule('/bad/pfns', view_func=bad_replicas_pfn_view, methods=['post', ])
    bad_replicas_dids_view = BadDIDs.as_view('add_bad_dids')
    bp.add_url_rule('/bad/dids', view_func=bad_replicas_dids_view, methods=['post', ])
    replicas_rse_view = ReplicasRSE.as_view('replicas_rse')
    bp.add_url_rule('/rse/<rse>', view_func=replicas_rse_view, methods=['get', ])

    bad_replicas_view = BadReplicas.as_view('bad_replicas')
    bp.add_url_rule('/bad', view_func=bad_replicas_view, methods=['post', ])

    quarantine_replicas_view = QuarantineReplicas.as_view('quarantine_replicas')
    bp.add_url_rule('/quarantine', view_func=quarantine_replicas_view, methods=['post', ])

    replicas_dids_view = ReplicasDIDs.as_view('replicas_dids')
    bp.add_url_rule('/dids', view_func=replicas_dids_view, methods=['post', ])
    dataset_replicas_view = DatasetReplicas.as_view('dataset_replicas')
    bp.add_url_rule('/<path:scope_name>/datasets', view_func=dataset_replicas_view, methods=['get', ])
    dataset_replicas_bulk_view = DatasetReplicasBulk.as_view('dataset_replicas_bulk')
    bp.add_url_rule('/datasets_bulk', view_func=dataset_replicas_bulk_view, methods=['post', ])
    dataset_replicas_vp_view = DatasetReplicasVP.as_view('dataset_replicas_vp')
    bp.add_url_rule('/<path:scope_name>', view_func=replicas_view, methods=['get', ])
    set_tombstone_view = Tombstone.as_view('set_tombstone')
    bp.add_url_rule('/tombstone', view_func=set_tombstone_view, methods=['post', ])

    if not with_doc:
        bp.add_url_rule('/list/', view_func=list_replicas_view, methods=['post', ])
        bp.add_url_rule('/suspicious/', view_func=suspicious_replicas_view, methods=['get', 'post'])
        bp.add_url_rule('/bad/states/', view_func=bad_replicas_states_view, methods=['get', ])
        bp.add_url_rule('/bad/summary/', view_func=bad_replicas_summary_view, methods=['get', ])
        bp.add_url_rule('/bad/pfns/', view_func=bad_replicas_pfn_view, methods=['post', ])
        bp.add_url_rule('/bad/dids/', view_func=bad_replicas_dids_view, methods=['post', ])
        bp.add_url_rule('/rse/<rse>/', view_func=replicas_rse_view, methods=['get', ])
        bp.add_url_rule('/bad/', view_func=bad_replicas_view, methods=['post', ])
        bp.add_url_rule('/dids/', view_func=replicas_dids_view, methods=['post', ])
        bp.add_url_rule('/datasets_bulk/', view_func=dataset_replicas_bulk_view, methods=['post', ])
        bp.add_url_rule('/<path:scope_name>/datasets_vp', view_func=dataset_replicas_vp_view, methods=['get', ])
        bp.add_url_rule('/<path:scope_name>/', view_func=replicas_view, methods=['get', ])
        bp.add_url_rule('/tombstone/', view_func=set_tombstone_view, methods=['post', ])
        bp.add_url_rule('/quarantine/', view_func=quarantine_replicas_view, methods=['post', ])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint(with_doc=True))
    return doc_app



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' remove unique constraint on requests '''

from alembic import context
from alembic.op import (create_foreign_key, create_index, drop_index,
                        drop_constraint, create_unique_constraint)


# Alembic revision identifiers
revision = '25821a8a45a3'
down_revision = '1803333ac20f'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        drop_constraint('REQUESTS_RSES_FK', 'requests', type_='foreignkey')
        drop_constraint('REQUESTS_DID_FK', 'requests', type_='foreignkey')
        drop_constraint('REQUESTS_SC_NA_RS_TY_UQ_IDX', 'requests', type_='unique')
        create_foreign_key('REQUESTS_RSES_FK', 'requests', 'rses', ['dest_rse_id'], ['id'])
        create_foreign_key('REQUESTS_DID_FK', 'requests', 'dids', ['scope', 'name'], ['scope', 'name'])
        create_index('REQUESTS_SCOPE_NAME_RSE_IDX', 'requests', ['scope', 'name', 'dest_rse_id', 'request_type'])


def downgrade():

    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        drop_constraint('REQUESTS_RSES_FK', 'requests', type_='foreignkey')
        drop_constraint('REQUESTS_DID_FK', 'requests', type_='foreignkey')
        drop_index('REQUESTS_SCOPE_NAME_RSE_IDX', 'requests')
        create_foreign_key('REQUESTS_RSES_FK', 'requests', 'rses', ['dest_rse_id'], ['id'])
        create_foreign_key('REQUESTS_DID_FK', 'requests', 'dids', ['scope', 'name'], ['scope', 'name'])
        create_unique_constraint('REQUESTS_SC_NA_RS_TY_UQ_IDX', 'requests', ['scope', 'name', 'dest_rse_id', 'request_type'])



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.api import permission
from rucio.common import exception
from rucio.common.config import convert_to_any_type
from rucio.core import config
from rucio.db.sqla.session import read_session, transactional_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

"""
ConfigParser compatible interface.

- File handling methods unnecessary.
- Convenience methods getint/getfloat/getboolean are superseded by auto-coercing get.
"""


@read_session
def sections(issuer=None, vo='def', *, session: "Session"):
    """
    Return a list of the sections available.

    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: ['section_name', ...]
    """

    kwargs = {'issuer': issuer}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_sections', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot retrieve sections' % issuer)
    return config.sections(session=session)


@transactional_session
def add_section(section, issuer=None, vo='def', *, session: "Session"):
    """
    Add a section to the configuration.

    :param section: The name of the section.
    :param issuer: The issuer account.
    :param session: The database session in use.
    :param vo: The VO to act on.
    """

    kwargs = {'issuer': issuer, 'section': section}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_add_section', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot add section %s' % (issuer, section))
    return config.add_section(section, session=session)


@read_session
def has_section(section, issuer=None, vo='def', *, session: "Session"):
    """
    Indicates whether the named section is present in the configuration.

    :param section: The name of the section.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: True/False
    """

    kwargs = {'issuer': issuer, 'section': section}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_has_section', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot check existence of section %s' % (issuer, section))
    return config.has_section(section, session=session)


@read_session
def options(section, issuer=None, vo='def', *, session: "Session"):
    """
    Returns a list of options available in the specified section.

    :param section: The name of the section.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: ['option', ...]
    """

    kwargs = {'issuer': issuer, 'section': section}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_options', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot retrieve options from section %s' % (issuer, section))
    return config.options(section, session=session)


@read_session
def has_option(section, option, issuer=None, vo='def', *, session: "Session"):
    """
    Check if the given section exists and contains the given option.

    :param section: The name of the section.
    :param option: The name of the option.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: True/False
    """

    kwargs = {'issuer': issuer, 'section': section, 'option': option}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_has_option', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot check existence of option %s from section %s' % (issuer, option, section))
    return config.has_option(section, option, session=session)


@read_session
def get(section, option, issuer=None, vo='def', *, session: "Session"):
    """
    Get an option value for the named section. Value can be auto-coerced to int, float, and bool; string otherwise.

    Caveat emptor: Strings, regardless the case, matching 'on'/off', 'true'/'false', 'yes'/'no' are converted to bool.
                   0/1 are converted to int, and not to bool.

    :param section: The name of the section.
    :param option: The name of the option.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: The auto-coerced value.
    """

    kwargs = {'issuer': issuer, 'section': section, 'option': option}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_get', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot retrieve option %s from section %s' % (issuer, option, section))
    return config.get(section, option, session=session, convert_type_fnc=convert_to_any_type)


@read_session
def items(section, issuer=None, vo='def', *, session: "Session"):
    """
    Return a list of (option, value) pairs for each option in the given section. Values are auto-coerced as in get().

    :param section: The name of the section.
    :param value: The content of the value.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: [('option', auto-coerced value), ...]
    """

    kwargs = {'issuer': issuer, 'section': section}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_items', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot retrieve options and values from section %s' % (issuer, section))
    return config.items(section, session=session, convert_type_fnc=convert_to_any_type)


@transactional_session
def set(section, option, value, issuer=None, vo='def', *, session: "Session"):
    """
    Set the given option to the specified value.

    :param section: The name of the section.
    :param option: The name of the option.
    :param value: The content of the value.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    """

    kwargs = {'issuer': issuer, 'section': section, 'option': option, 'value': value}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_set', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot set option %s to %s in section %s' % (issuer, option, value, section))
    return config.set(section, option, value, session=session)


@transactional_session
def remove_section(section, issuer=None, vo='def', *, session: "Session"):
    """
    Remove the specified option from the specified section.

    :param section: The name of the section.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: True/False.
    """

    kwargs = {'issuer': issuer, 'section': section}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_remove_section', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot remove section %s' % (issuer, section))
    return config.remove_section(section, session=session)


@transactional_session
def remove_option(section, option, issuer=None, vo='def', *, session: "Session"):
    """
    Remove the specified section from the configuration.

    :param section: The name of the section.
    :param option: The name of the option.
    :param issuer: The issuer account.
    :param vo: The VO to act on.
    :param session: The database session in use.
    :returns: True/False
    """

    kwargs = {'issuer': issuer, 'section': section, 'option': option}
    if not permission.has_permission(issuer=issuer, vo=vo, action='config_remove_option', kwargs=kwargs, session=session):
        raise exception.AccessDenied('%s cannot remove option %s from section %s' % (issuer, option, section))
    return config.remove_option(section, option, session=session)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Expiring Dataset Cache
"""
from uuid import uuid4

from redis import StrictRedis


class ExpiringDatasetCache(object):
    """
    Cache with expiring values to keep track of recently created replicas.
    """
    def __init__(self, redis_host, redis_port, timeout=1, prefix='expiring_did_cache'):
        self._redis = StrictRedis(host=redis_host, port=redis_port)
        self._prefix = prefix + '_' + str(uuid4()).split('-')[0]
        self._timeout = timeout

    def add_dataset(self, dataset):
        """ Adds a datasets to cache with lifetime """
        key = ':'.join((self._prefix, dataset))
        self._redis.set(key, 1)
        self._redis.expire(key, self._timeout)

    def check_dataset(self, dataset):
        """ Checks if dataset is still in cache """
        key = ':'.join((self._prefix, dataset))
        if self._redis.get(key) is None:
            return False

        return True



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" state idx non functional """

from alembic import context
from alembic.op import create_index, drop_index, execute

# Alembic revision identifiers
revision = 'a6eb23955c28'
down_revision = 'fb28a95fe288'


def upgrade():
    """
    Upgrade the database to this revision
    """

    schema = context.get_context().version_table_schema + '.' if context.get_context().version_table_schema else ''
    if context.get_context().dialect.name in ['oracle', 'postgresql']:
        execute(f'ALTER INDEX {schema}"RULES_STUCKSTATE_IDX" RENAME TO "RULES_STATE_IDX"')
    elif context.get_context().dialect.name == 'mysql':
        execute(f'ALTER TABLE {schema}rules RENAME INDEX RULES_STUCKSTATE_IDX TO RULES_STATE_IDX')
    elif context.get_context().dialect.name == 'sqlite':
        create_index('RULES_STATE_IDX', 'rules', ['state'])
        drop_index('RULES_STUCKSTATE_IDX', 'rules')


def downgrade():
    """
    Downgrade the database to the previous revision
    """

    schema = context.get_context().version_table_schema + '.' if context.get_context().version_table_schema else ''
    if context.get_context().dialect.name in ['oracle', 'postgresql']:
        execute(f'ALTER INDEX {schema}"RULES_STATE_IDX" RENAME TO "RULES_STUCKSTATE_IDX"')
    elif context.get_context().dialect.name == 'mysql':
        execute(f'ALTER TABLE {schema}rules RENAME INDEX RULES_STATE_IDX TO RULES_STUCKSTATE_IDX')
    elif context.get_context().dialect.name == 'sqlite':
        create_index('RULES_STUCKSTATE_IDX', 'rules', ['state'])
        drop_index('RULES_STATE_IDX', 'rules')



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import loads
from urllib.parse import quote_plus

from requests.status_codes import codes

from rucio.client.baseclient import BaseClient
from rucio.client.baseclient import choice
from rucio.common.utils import build_url


class FileClient(BaseClient):
    """Dataset client class for working with dataset"""

    BASEURL = 'files'

    def list_file_replicas(self, scope, lfn):
        """
        List file replicas.

        :param scope: the scope.
        :param lfn: the lfn.

        :return: List of replicas.
        """
        path = '/'.join([self.BASEURL, quote_plus(scope), quote_plus(lfn), 'rses'])
        url = build_url(choice(self.list_hosts), path=path)

        r = self._send_request(url, type_='GET')

        if r.status_code == codes.ok:
            rses = loads(r.text)
            return rses
        else:
            print(r.status_code)
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import hashlib
import random
import re
import sys
import traceback
from base64 import b64decode
from typing import TYPE_CHECKING

import paramiko
from dogpile.cache import make_region
from dogpile.cache.api import NO_VALUE
from sqlalchemy import delete, null, or_, select

from rucio.common.cache import make_region_memcached
from rucio.common.config import config_get_bool
from rucio.common.exception import CannotAuthenticate, RucioException
from rucio.common.utils import chunks, generate_uuid, date_to_str
from rucio.core.account import account_exists
from rucio.core.oidc import validate_jwt
from rucio.db.sqla import filter_thread_work
from rucio.db.sqla import models
from rucio.db.sqla.constants import IdentityType
from rucio.db.sqla.session import read_session, transactional_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session
    from typing import Any, Union


def strip_x509_proxy_attributes(dn: str) -> str:
    """Strip X509 proxy attributes from a DN.

    When an X509 VOMS proxy certificate is produced, an additional Common Name
    attribute is added to the subject of the original certificate.  Its value
    can take different forms.  For proxy versions 3 and later (the default), the
    value is a numeric.  For previous versions, the value is exclusively one of
    'limited proxy' or 'proxy', depending on how it was produced (the most
    trustworthy documentation on this seems to be the VOMS source code itself;
    refer to the file sslutils.c).  Note that this addition might happen more
    than once (e.g. if a limited proxy is used to produce a full proxy).

    By default, the Apache server will return the DN in an RFC-compliant format,
    which can look like this:
        CN=John Doe,OU=Users,DC=example,DC=com
    However, in case the LegacyDNStringFormat of mod_ssl is enabled, then it can
    look like this instead:
        /DC=com/DC=example/OU=Users/CN=John Doe
    In the first case, the Common Name attributes added by VOMS are prepended,
    whereas in the second case, they are appended.

    The motivation for stripping these attributes is to avoid having to store
    multiple DNs in the database (as different identities).
    """
    if dn.startswith('/'):
        regexp = r'(/CN=(limited proxy|proxy|[0-9]+))+$'
    else:
        regexp = r'^(CN=(limited proxy|proxy|[0-9]+),)+'

    return re.sub(regexp, '', dn)


def token_key_generator(namespace, fni, **kwargs):
    """ :returns: generate key function """
    def generate_key(token, *, session: "Session"):
        """ :returns: token """
        return token
    return generate_key


if config_get_bool('cache', 'use_external_cache_for_auth_tokens', default=False):
    TOKENREGION = make_region_memcached(expiration_time=900, function_key_generator=token_key_generator)
else:
    TOKENREGION = make_region(function_key_generator=token_key_generator).configure('dogpile.cache.memory', expiration_time=900)


@transactional_session
def get_auth_token_user_pass(account, username, password, appid, ip=None, *, session: "Session"):
    """
    Authenticate a Rucio account temporarily via username and password.

    The token lifetime is 1 hour.

    :param account: Account identifier as a string.
    :param username: Username as a string.
    :param password: SHA1 hash of the password as a string.
    :param appid: The application identifier as a string.
    :param ip: IP address of the client a a string.
    :param session: The database session in use.

    :returns: A dict with token and expires_at entries.
    """

    # Make sure the account exists
    if not account_exists(account, session=session):
        return None

    query = select(
        models.Identity
    ).where(
        models.Identity.identity == username,
        models.Identity.identity_type == IdentityType.USERPASS
    )
    result = session.execute(query).scalar()

    db_salt = result['salt']
    db_password = result['password']

    salted_password = db_salt + password.encode()
    if db_password != hashlib.sha256(salted_password).hexdigest():
        return None

    # get account identifier
    query = select(
        models.IdentityAccountAssociation
    ).where(
        models.IdentityAccountAssociation.identity == username,
        models.IdentityAccountAssociation.identity_type == IdentityType.USERPASS,
        models.IdentityAccountAssociation.account == account
    )
    result = session.execute(query).scalar()

    db_account = result['account']

    # remove expired tokens
    __delete_expired_tokens_account(account=account, session=session)

    # create new rucio-auth-token for account
    tuid = generate_uuid()  # NOQA
    token = f'{account}-{username}-{appid}-{tuid}'
    new_token = models.Token(account=db_account, identity=username, token=token, ip=ip)
    new_token.save(session=session)

    return token_dictionary(new_token)


@transactional_session
def get_auth_token_x509(account, dn, appid, ip=None, *, session: "Session"):
    """
    Authenticate a Rucio account temporarily via an x509 certificate.

    The token lifetime is 1 hour.

    :param account: Account identifier as a string.
    :param dn: Client certificate distinguished name string, as extracted by Apache/mod_ssl.
    :param id: The application identifier as a string.
    :param ipaddr: IP address of the client as a string.
    :param session: The database session in use.

    :returns: A dict with token and expires_at entries.
    """

    # Make sure the account exists
    if not account_exists(account, session=session):
        return None

    # remove expired tokens
    __delete_expired_tokens_account(account=account, session=session)

    # create new rucio-auth-token for account
    tuid = generate_uuid()  # NOQA
    token = f'{account}-{dn}-{appid}-{tuid}'
    new_token = models.Token(account=account, identity=dn, token=token, ip=ip)
    new_token.save(session=session)

    return token_dictionary(new_token)


@transactional_session
def get_auth_token_gss(account, gsstoken, appid, ip=None, *, session: "Session"):
    """
    Authenticate a Rucio account temporarily via a GSS token.

    The token lifetime is 1 hour.

    :param account: Account identifier as a string.
    :param gsscred: GSS principal@REALM as a string.
    :param appid: The application identifier as a string.
    :param ip: IP address of the client as a string.
    :param session: The database session in use.

    :returns: A dict with token and expires_at entries.
    """

    # Make sure the account exists
    if not account_exists(account, session=session):
        return None

    # remove expired tokens
    __delete_expired_tokens_account(account=account, session=session)

    # create new rucio-auth-token for account
    tuid = generate_uuid()  # NOQA
    token = f'{account}-{gsstoken}-{appid}-{tuid}'
    new_token = models.Token(account=account, token=token, ip=ip)
    new_token.save(session=session)

    return token_dictionary(new_token)


@transactional_session
def get_auth_token_ssh(account, signature, appid, ip=None, *, session: "Session"):
    """
    Authenticate a Rucio account temporarily via SSH key exchange.

    The token lifetime is 1 hour.

    :param account: Account identifier as a string.
    :param signature: Response to server challenge signed with SSH private key as string.
    :param appid: The application identifier as a string.
    :param ip: IP address of the client as a string.
    :param session: The database session in use.

    :returns: A dict with token and expires_at entries.
    """
    if not isinstance(signature, bytes):
        signature = signature.encode()

    # Make sure the account exists
    if not account_exists(account, session=session):
        return None

    # get all active challenge tokens for the requested account
    query = select(
        models.Token
    ).where(
        models.Token.expired_at >= datetime.datetime.utcnow(),
        models.Token.account == account,
        models.Token.token.like('challenge-%')
    )
    active_challenge_tokens = session.execute(query).scalars().all()

    # get all identities for the requested account
    query = select(
        models.IdentityAccountAssociation
    ).where(
        models.IdentityAccountAssociation.identity_type == IdentityType.SSH,
        models.IdentityAccountAssociation.account == account
    )
    identities = session.execute(query).scalars().all()

    # no challenge tokens found
    if not active_challenge_tokens:
        return None

    # try all available SSH identities for the account with the provided signature
    match = False
    for identity in identities:
        data = identity['identity'].split()[1]
        data += '=' * ((4 - len(data) % 4) % 4)  # adding required padding
        pub_k = paramiko.RSAKey(data=b64decode(data))
        for challenge_token in active_challenge_tokens:
            if pub_k.verify_ssh_sig(str(challenge_token['token']).encode(),
                                    paramiko.Message(signature)):
                match = True
                break
        if match:
            break

    if not match:
        return None

    # remove expired tokens
    __delete_expired_tokens_account(account=account, session=session)

    # create new rucio-auth-token for account
    tuid = generate_uuid()  # NOQA
    token = f'{account}-ssh:pubkey-{appid}-{tuid}'
    new_token = models.Token(account=account, token=token, ip=ip)
    new_token.save(session=session)

    return token_dictionary(new_token)


@transactional_session
def get_ssh_challenge_token(account, appid, ip=None, *, session: "Session"):
    """
    Prepare a challenge token for subsequent SSH public key authentication.

    The challenge lifetime is fixed to 10 seconds.

    :param account: Account identifier as a string.
    :param appid: The application identifier as a string.
    :param ip: IP address of the client as a string.

    :returns: A dict with token and expires_at entries.
    """

    # Make sure the account exists
    if not account_exists(account, session=session):
        return None

    # Cryptographically secure random number.
    # This requires a /dev/urandom like device from the OS
    rng = random.SystemRandom()
    crypto_rand = rng.randint(0, sys.maxsize)

    # give the client 10 seconds max to sign the challenge token
    expiration = datetime.datetime.utcnow() + datetime.timedelta(seconds=10)
    expiration_unix = expiration.strftime("%s")

    challenge_token = f'challenge-{crypto_rand}-{account}-{expiration_unix}'

    new_challenge_token = models.Token(account=account, token=challenge_token, ip=ip,
                                       expired_at=expiration)
    new_challenge_token.save(session=session)

    return token_dictionary(new_challenge_token)


@transactional_session
def get_auth_token_saml(account, saml_nameid, appid, ip=None, *, session: "Session"):
    """
    Authenticate a Rucio account temporarily via SAML.

    The token lifetime is 1 hour.

    :param account: Account identifier as a string.
    :param saml_nameid: SAML NameID of the client.
    :param appid: The application identifier as a string.
    :param ip: IP address of the client a a string.
    :param session: The database session in use.

    :returns: A dict with token and expires_at entries.
    """

    # Make sure the account exists
    if not account_exists(account, session=session):
        return None

    # remove expired tokens
    __delete_expired_tokens_account(account=account, session=session)

    tuid = generate_uuid()  # NOQA
    token = f'{account}-{saml_nameid}-{appid}-{tuid}'
    new_token = models.Token(account=account, identity=saml_nameid, token=token, ip=ip)
    new_token.save(session=session)

    return token_dictionary(new_token)


@transactional_session
def redirect_auth_oidc(auth_code, fetchtoken=False, *, session: "Session"):
    """
    Finds the Authentication URL in the Rucio DB oauth_requests table
    and redirects user's browser to this URL.

    :param auth_code: Rucio assigned code to redirect
                      authorization securely to IdP via Rucio Auth server through a browser.
    :param fetchtoken: If True, valid token temporarily saved in the oauth_requests table
                       will be returned. If False, redirection URL is returned.
    :param session: The database session in use.

    :returns: result of the query (authorization URL or a
              token if a user asks with the correct code) or None.
              Exception thrown in case of an unexpected crash.

    """
    try:
        query = select(
            models.OAuthRequest.redirect_msg
        ).where(
            models.OAuthRequest.access_msg == auth_code
        )
        redirect_result = session.execute(query).scalar()

        if not redirect_result:
            return None

        if 'http' not in redirect_result and fetchtoken:
            # in this case the function check if the value is a valid token
            vdict = validate_auth_token(redirect_result, session=session)
            if vdict:
                return redirect_result
            return None
        elif 'http' in redirect_result and not fetchtoken:
            # return redirection URL
            return redirect_result
    except:
        raise CannotAuthenticate(traceback.format_exc())


@transactional_session
def delete_expired_tokens(total_workers, worker_number, limit=1000, *, session: "Session"):
    """
    Delete expired tokens.

    :param total_workers:      Number of total workers.
    :param worker_number:      id of the executing worker.
    :param limit:              Maximum number of tokens to delete.
    :param session:            Database session in use.

    :returns: number of deleted rows
    """

    # get expired tokens
    try:
        # delete all expired tokens except tokens which have refresh token that is still valid
        query = select(
            models.Token.token
        ).where(
            models.Token.expired_at <= datetime.datetime.utcnow(),
            or_(
                models.Token.refresh_expired_at == null(),
                models.Token.refresh_expired_at <= datetime.datetime.utcnow()
            )
        ).order_by(
            models.Token.expired_at
        )

        query = filter_thread_work(session=session, query=query, total_threads=total_workers, thread_id=worker_number, hash_variable='token')

        # limiting the number of tokens deleted at once
        query = query.limit(limit)
        # Oracle does not support chaining order_by(), limit(), and
        # with_for_update(). Use a nested query to overcome this.
        if session.bind.dialect.name == 'oracle':
            query = select(
                models.Token.token
            ).where(
                models.Token.token.in_(query)
            ).with_for_update(
                skip_locked=True
            )
        else:
            query = query.with_for_update(skip_locked=True)
        # remove expired tokens
        deleted_tokens = 0
        for tokens in session.execute(query).scalars().partitions(10):
            query = delete(
                models.Token
            ).where(
                models.Token.token.in_(tokens)
            )
            deleted_tokens += session.execute(query).rowcount

    except Exception as error:
        raise RucioException(error.args)

    return deleted_tokens


@read_session
def query_token(token, *, session: "Session"):
    """
    Validate an authentication token using the database. This method will only be called
    if no entry could be found in the according cache.

    :param token: Authentication token as a variable-length string.
    :param session: The database session in use.

    :returns: dictionary { account: <account name>,
                           identity: <identity>,
                           lifetime: <token lifetime>,
                           audience: <audience>,
                           authz_scope: <authz_scope> }
              if successful, None otherwise.
    """
    # Query the DB to validate token
    query = select(
        models.Token.account,
        models.Token.identity,
        models.Token.expired_at.label('lifetime'),
        models.Token.audience,
        models.Token.oidc_scope.label('authz_scope')
    ).where(
        models.Token.token == token,
        models.Token.expired_at > datetime.datetime.utcnow()
    )
    result = session.execute(query).first()
    if result:
        return result._asdict()
    return None


@transactional_session
def validate_auth_token(token: str, *, session: "Session") -> "dict[str, Any]":
    """
    Validate an authentication token.

    :param token: Authentication token as a variable-length string.

    :returns: dictionary { account: <account name>,
                           identity: <identity>,
                           lifetime: <token lifetime>,
                           audience: <audience>,
                           authz_scope: <authz_scope> }
              if successful
    :raises: CannotAuthenticate if unsuccessful
    """
    if not token:
        raise CannotAuthenticate("No token was passed!")

    # Be gentle with bash variables, there can be whitespace
    token = token.strip()
    cache_key = token.replace(' ', '')

    # Check if token ca be found in cache region
    value: "Union[NO_VALUE, dict[str, Any]]" = TOKENREGION.get(cache_key)
    if value is NO_VALUE:  # no cached entry found
        value = query_token(token, session=session)
        if not value:
            # identify JWT access token and validte
            # & save it in Rucio if scope and audience are correct
            if len(token.split(".")) == 3:
                value = validate_jwt(token, session=session)
            else:
                raise CannotAuthenticate(traceback.format_exc())
        # save token in the cache
        TOKENREGION.set(cache_key, value)
    lifetime = value.get('lifetime', datetime.datetime(1970, 1, 1))
    if lifetime < datetime.datetime.utcnow():  # check if expired
        TOKENREGION.delete(cache_key)
        raise CannotAuthenticate(f"Token found but expired since {date_to_str(lifetime)}.")
    return value


def token_dictionary(token: models.Token):
    return {'token': token.token, 'expires_at': token.expired_at}


@transactional_session
def __delete_expired_tokens_account(account, *, session: "Session"):
    """"
    Deletes expired tokens from the database.

    :param account: Account to delete expired tokens.
    :param session: The database session in use.
    """
    select_query = select(
        models.Token.token
    ).where(
        models.Token.expired_at < datetime.datetime.utcnow(),
        models.Token.account == account
    ).with_for_update(
        skip_locked=True
    )
    tokens = session.execute(select_query).scalars().all()

    for chunk in chunks(tokens, 100):
        delete_query = delete(
            models.Token
        ).prefix_with(
            "/*+ INDEX(TOKENS_ACCOUNT_EXPIRED_AT_IDX) */"
        ).where(
            models.Token.token.in_(chunk)
        )
        session.execute(delete_query)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from typing import TYPE_CHECKING

from sqlalchemy import and_, or_, exists, update, insert
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import select

from rucio.common import exception
from rucio.core.rse import get_rse_name
from rucio.db.sqla import models
from rucio.db.sqla.constants import ReplicaState
from rucio.db.sqla.session import transactional_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


@transactional_session
def add_volatile_replicas(rse_id, replicas, *, session: "Session"):
    """
    Bulk add volatile replicas.

    :param rse_id: the rse id.
    :param replicas: the list of volatile replicas.
    :param session: The database session in use.
    :returns: True is successful.
    """
    # first check that the rse is a volatile one
    try:
        session.query(models.RSE).filter_by(id=rse_id, volatile=True).one()
    except NoResultFound:
        raise exception.UnsupportedOperation('No volatile rse found for %s !'
                                             % get_rse_name(rse_id=rse_id, session=session))

    file_clause, replica_clause = [], []
    for replica in replicas:
        file_clause.append(and_(models.DataIdentifier.scope == replica['scope'],
                                models.DataIdentifier.name == replica['name'],
                                ~exists(select(1).prefix_with("/*+ INDEX(REPLICAS REPLICAS_PK) */", dialect='oracle'))
                                .where(and_(models.RSEFileAssociation.scope == replica['scope'],
                                            models.RSEFileAssociation.name == replica['name'],
                                            models.RSEFileAssociation.rse_id == rse_id))))
        replica_clause.append(and_(models.RSEFileAssociation.scope == replica['scope'],
                                   models.RSEFileAssociation.name == replica['name'],
                                   models.RSEFileAssociation.rse_id == rse_id))
    if replica_clause:
        now = datetime.utcnow()
        stmt = update(models.RSEFileAssociation).\
            prefix_with("/*+ index(REPLICAS REPLICAS_PK) */", dialect='oracle').\
            where(or_(*replica_clause)).\
            execution_options(synchronize_session=False).\
            values(updated_at=now, tombstone=now)
        session.execute(stmt)

    if file_clause:
        file_query = session.query(models.DataIdentifier.scope,
                                   models.DataIdentifier.name,
                                   models.DataIdentifier.bytes,
                                   models.DataIdentifier.md5,
                                   models.DataIdentifier.adler32).\
            filter(or_(*file_clause))

        new_replicas = [
            {
                'rse_id': rse_id,
                'adler32': adler32,
                'state': ReplicaState.AVAILABLE,
                'scope': scope,
                'name': name,
                'lock_cnt': 0,
                'tombstone': datetime.utcnow(),
                'bytes': bytes_,
                'md5': md5
            }
            for scope, name, bytes_, md5, adler32 in file_query
        ]
        if new_replicas:
            session.execute(insert(models.RSEFileAssociation), new_replicas)


@transactional_session
def delete_volatile_replicas(rse_id, replicas, *, session: "Session"):
    """
    Bulk delete volatile replicas.

    :param rse_id: the rse id.
    :param replicas: the list of volatile replicas.
    :param session: The database session in use.
    :returns: True is successful.
    """
    # first check that the rse is a volatile one
    try:
        session.query(models.RSE).filter_by(id=rse_id, volatile=True).one()
    except NoResultFound:
        raise exception.UnsupportedOperation('No volatile rse found for %s !'
                                             % get_rse_name(rse_id=rse_id, session=session))

    conditions = []
    for replica in replicas:
        conditions.append(and_(models.RSEFileAssociation.scope == replica['scope'],
                               models.RSEFileAssociation.name == replica['name']))

    if conditions:
        session.query(models.RSEFileAssociation).\
            filter(models.RSEFileAssociation.rse_id == rse_id).\
            filter(or_(*conditions)).\
            delete(synchronize_session=False)



# -*- coding: utf-8 -*-
# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import datetime
import hmac
import time
from hashlib import sha1
from urllib.parse import urlparse, urlencode

import boto3
from botocore.client import Config
from dogpile.cache.api import NO_VALUE
from google.oauth2.service_account import Credentials

from rucio.common.cache import make_region_memcached
from rucio.common.config import config_get, get_rse_credentials
from rucio.common.exception import UnsupportedOperation
from rucio.core.monitor import MetricManager
from rucio.core.rse import get_rse_attribute

CREDS_GCS = None

REGION = make_region_memcached(expiration_time=900)
METRICS = MetricManager(module=__name__)


def get_signed_url(rse_id: str, service: str, operation: str, url: str, lifetime=600) -> str:
    """
    Get a signed URL for a particular service and operation.

    The signed URL will be valid for 1 hour but can be overriden.

    :param rse_id: The ID of the RSE that the URL points to.
    :param service: The service to authorise, either 'gcs', 's3' or 'swift'.
    :param operation: The operation to sign, either 'read', 'write', or 'delete'.
    :param url: The URL to sign.
    :param lifetime: Lifetime of the signed URL in seconds.
    :returns: Signed URL as a variable-length string.
    """

    global CREDS_GCS

    if service not in ['gcs', 's3', 'swift']:
        raise UnsupportedOperation('Service must be "gcs", "s3" or "swift"')

    if operation not in ['read', 'write', 'delete']:
        raise UnsupportedOperation('Operation must be "read", "write", or "delete"')

    if url is None or url == '':
        raise UnsupportedOperation('URL must not be empty')

    if lifetime:
        if not isinstance(lifetime, int):
            try:
                lifetime = int(lifetime)
            except:
                raise UnsupportedOperation('Lifetime must be convertible to numeric.')

    if service == 'gcs':
        if not CREDS_GCS:
            CREDS_GCS = Credentials.from_service_account_file(config_get('credentials', 'gcs',
                                                                         raise_exception=False,
                                                                         default='/opt/rucio/etc/google-cloud-storage-test.json'))
        components = urlparse(url)
        host = components.netloc

        # select the correct operation
        operations = {'read': 'GET', 'write': 'PUT', 'delete': 'DELETE'}
        operation = operations[operation]

        # special case to test signature, force epoch time
        if lifetime is None:
            lifetime = 0
        else:
            # GCS is timezone-sensitive, don't use UTC
            # has to be converted to Unixtime
            lifetime = datetime.datetime.now() + datetime.timedelta(seconds=lifetime)
            lifetime = int(time.mktime(lifetime.timetuple()))

        # sign the path only
        path = components.path

        # assemble message to sign
        to_sign = "%s\n\n\n%s\n%s" % (operation, lifetime, path)

        # create URL-capable signature
        # first character is always a '=', remove it
        signature = urlencode({'': base64.b64encode(CREDS_GCS.sign_bytes(to_sign))})[1:]

        # assemble final signed URL
        signed_url = (
            f'https://{host}{path}'
            f'?GoogleAccessId={CREDS_GCS.service_account_email}'
            f'&Expires={lifetime}&Signature={signature}'
        )

    elif service == 's3':

        # get RSE S3 URL style (path or host)
        # path-style: https://s3.region-code.amazonaws.com/bucket-name/key-name
        # host-style: https://bucket-name.s3.region-code.amazonaws.com/key-name
        s3_url_style = get_rse_attribute(rse_id, 's3_url_style')

        # no S3 URL style specified, assume path-style
        if s3_url_style is None:
            s3_url_style = "path"

        # split URL to get hostname, bucket and key
        components = urlparse(url)
        host = components.netloc
        pathcomponents = components.path.split('/')
        if s3_url_style == "path":
            if len(pathcomponents) < 3:
                raise UnsupportedOperation('Not a valid Path-Style S3 URL')
            bucket = pathcomponents[1]
            key = '/'.join(pathcomponents[2:])
        elif s3_url_style == "host":
            hostcomponents = host.split('.')
            bucket = hostcomponents[0]
            if len(pathcomponents) < 2:
                raise UnsupportedOperation('Not a valid Host-Style S3 URL')
            key = '/'.join(pathcomponents[1:])
        else:
            raise UnsupportedOperation('Not a valid RSE S3 URL style (allowed values: path|host)')

        # remove port number from host if present
        colon = host.find(':')
        port = '443'
        if colon >= 0:
            port = host[colon + 1:]
            host = host[:colon]

        # look up in RSE account configuration by RSE ID
        cred_name = rse_id
        cred = REGION.get('s3-%s' % cred_name)
        if cred is NO_VALUE:
            rse_cred = get_rse_credentials()
            cred = rse_cred.get(cred_name)
            REGION.set('s3-%s' % cred_name, cred)
        access_key = cred['access_key']
        secret_key = cred['secret_key']
        signature_version = cred['signature_version']
        region_name = cred['region']

        if operation == 'read':
            s3op = 'get_object'
        elif operation == 'write':
            s3op = 'put_object'
        else:
            s3op = 'delete_object'

        with METRICS.timer('signs3'):

            if s3_url_style == "host":
                s3_url_style = "virtual"

            s3 = boto3.client(service_name='s3',
                              endpoint_url=f'https://{host}:{port}',
                              aws_access_key_id=access_key,
                              aws_secret_access_key=secret_key,
                              config=Config(signature_version=signature_version,
                                            region_name=region_name,
                                            s3={"addressing_style": s3_url_style}))

            signed_url: str = s3.generate_presigned_url(
                s3op, Params={'Bucket': bucket, 'Key': key}, ExpiresIn=lifetime)

    else:  # service == 'swift'
        # split URL to get hostname and path
        components = urlparse(url)
        host = components.netloc

        # remove port number from host if present
        colon = host.find(':')
        if colon >= 0:
            host = host[:colon]

        # use RSE ID to look up key
        cred_name = rse_id

        # look up tempurl signing key
        cred = REGION.get('swift-%s' % cred_name)
        if cred is NO_VALUE:
            rse_cred = get_rse_credentials()
            cred = rse_cred.get(cred_name)
            REGION.set('swift-%s' % cred_name, cred)
        tempurl_key = cred['tempurl_key']

        if operation == 'read':
            swiftop = 'GET'
        elif operation == 'write':
            swiftop = 'PUT'
        else:
            swiftop = 'DELETE'

        expires = int(time.time() + lifetime)

        # create signed URL
        with METRICS.timer('signswift'):
            hmac_body = '%s\n%s\n%s' % (swiftop, expires, components.path)
            # Python 3 hmac only accepts bytes or bytearray
            sig = hmac.new(bytearray(tempurl_key, 'utf-8'), bytearray(hmac_body, 'utf-8'), sha1).hexdigest()
            signed_url = f'https://{host}{components.path}?temp_url_sig={sig}&temp_url_expires={expires}'

    return signed_url
